"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var ts = require("typescript");
var Lint = require("tslint");
var Rule = (function (_super) {
    __extends(Rule, _super);
    function Rule() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    Rule.prototype.apply = function (sourceFile) {
        return this.applyWithFunction(sourceFile, walk, parseOptions(this.ruleArguments));
    };
    return Rule;
}(Lint.Rules.AbstractRule));
Rule.FAILURE_STRING = "Only ReadonlyArray allowed.";
exports.Rule = Rule;
var OPTION_IGNORE_LOCAL = "ignore-local";
var OPTION_IGNORE_PREFIX = "ignore-prefix";
function createInvalidNode(node, replacement) {
    return { node: node, replacement: replacement };
}
function parseOptions(options) {
    var ignoreLocal = options.indexOf(OPTION_IGNORE_LOCAL) !== -1;
    var ignorePrefix;
    for (var _i = 0, options_1 = options; _i < options_1.length; _i++) {
        var o = options_1[_i];
        if (typeof o === "object" && o[OPTION_IGNORE_PREFIX] !== null) {
            ignorePrefix = o[OPTION_IGNORE_PREFIX];
            break;
        }
    }
    return { ignoreLocal: ignoreLocal, ignorePrefix: ignorePrefix };
}
function walk(ctx) {
    return ts.forEachChild(ctx.sourceFile, cb);
    function cb(node) {
        // Skip checking in functions if ignore-local is set
        if (ctx.options.ignoreLocal && (node.kind === ts.SyntaxKind.FunctionDeclaration
            || node.kind === ts.SyntaxKind.ArrowFunction || node.kind === ts.SyntaxKind.FunctionExpression)) {
            // We still need to check the parameters and return type
            var functionNode = node; //tslint:disable-line
            var invalidNodes = checkIgnoreLocalFunctionNode(functionNode, ctx);
            invalidNodes.forEach(function (n) { return reportInvalidNode(n, ctx); });
            // Now skip this whole branch
            return;
        }
        // Check the node
        reportInvalidNode(checkNode(node, ctx), ctx);
        // Use return becuase performance hints docs say it optimizes the function using tail-call recursion
        return ts.forEachChild(node, cb);
    }
}
function checkNode(node, ctx) {
    return checkArrayTypeOrReference(node, ctx) || checkVariableOrParameterImplicitType(node, ctx);
}
function reportInvalidNode(invalidNode, ctx) {
    if (invalidNode) {
        ctx.addFailureAtNode(invalidNode.node, Rule.FAILURE_STRING, invalidNode.replacement);
    }
}
function checkIgnoreLocalFunctionNode(functionNode, ctx) {
    var invalidNodes = [];
    // Check either the parameter's explicit type if it has one, or itself for implict type
    for (var _i = 0, _a = functionNode.parameters.map(function (p) { return p.type ? p.type : p; }); _i < _a.length; _i++) {
        var n = _a[_i];
        var invalidNode = checkNode(n, ctx);
        if (invalidNode) {
            invalidNodes.push(invalidNode);
        }
    }
    // Check the return type
    if (functionNode.type) {
        var invalidNode = checkNode(functionNode.type, ctx);
        if (invalidNode) {
            invalidNodes.push(invalidNode);
        }
    }
    return invalidNodes;
}
function checkArrayTypeOrReference(node, ctx) {
    // We need to check both shorthand syntax "number[]" and type reference "Array<number>"
    if (node.kind === ts.SyntaxKind.ArrayType
        || (node.kind === ts.SyntaxKind.TypeReference && node.typeName.getText(ctx.sourceFile) === "Array")) {
        if (node.parent && shouldIgnorePrefix(node.parent, ctx.options, ctx.sourceFile)) {
            return undefined;
        }
        var typeArgument = "T";
        if (node.kind === ts.SyntaxKind.ArrayType) {
            var typeNode = node;
            typeArgument = typeNode.elementType.getFullText(ctx.sourceFile).trim();
        }
        else if (node.kind === ts.SyntaxKind.TypeReference) {
            var typeNode = node;
            if (typeNode.typeArguments) {
                typeArgument = typeNode.typeArguments[0].getFullText(ctx.sourceFile).trim();
            }
        }
        var length_1 = node.getWidth(ctx.sourceFile);
        return createInvalidNode(node, new Lint.Replacement(node.end - length_1, length_1, "ReadonlyArray<" + typeArgument + ">"));
    }
    return undefined;
}
function checkVariableOrParameterImplicitType(node, ctx) {
    if (node.kind === ts.SyntaxKind.VariableDeclaration || node.kind === ts.SyntaxKind.Parameter) {
        // The initializer is used to set and implicit type
        var varOrParamNode = node;
        if (shouldIgnorePrefix(node, ctx.options, ctx.sourceFile)) {
            return undefined;
        }
        if (!varOrParamNode.type) {
            if (varOrParamNode.initializer && varOrParamNode.initializer.kind === ts.SyntaxKind.ArrayLiteralExpression) {
                var length_2 = varOrParamNode.name.getWidth(ctx.sourceFile);
                var nameText = varOrParamNode.name.getText(ctx.sourceFile);
                var typeArgument = "any";
                // Not sure it is a good idea to guess what the element types are...
                // const arrayLiteralNode = varOrParamNode.initializer as ts.ArrayLiteralExpression;
                // if (arrayLiteralNode.elements.length > 0) {
                //   const element = arrayLiteralNode.elements[0];
                //   if (element.kind === ts.SyntaxKind.NumericLiteral) {
                //     typeArgument = "number";
                //   } else if (element.kind === ts.SyntaxKind.StringLiteral) {
                //     typeArgument = "string";
                //   } else if (element.kind === ts.SyntaxKind.TrueKeyword || element.kind === ts.SyntaxKind.FalseKeyword) {
                //     typeArgument = "boolean";
                //   }
                // }
                return createInvalidNode(varOrParamNode.name, new Lint.Replacement(varOrParamNode.name.end - length_2, length_2, nameText + ": ReadonlyArray<" + typeArgument + ">"));
            }
        }
    }
    return undefined;
}
function shouldIgnorePrefix(node, options, sourceFile) {
    // Check ignore-prefix for VariableDeclaration, PropertySignature, TypeAliasDeclaration, Parameter
    if (options.ignorePrefix) {
        if (node && (node.kind === ts.SyntaxKind.VariableDeclaration
            || node.kind === ts.SyntaxKind.Parameter
            || node.kind === ts.SyntaxKind.PropertySignature
            || node.kind === ts.SyntaxKind.TypeAliasDeclaration)) {
            var variableDeclarationNode = node;
            if (variableDeclarationNode.name.getText(sourceFile).substr(0, options.ignorePrefix.length) === options.ignorePrefix) {
                return true;
            }
        }
    }
    return false;
}
